/*
    This is a Second Extended File System driver for Windows NT.

    Written by John Newbigin jn@it.swin.edu.au

	Thanks to Bo Branten for sample code and ntifs.h

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Header: /home/itig/cvsroot/ext2ifs/inc/ext2fs.h,v 1.15 2001/10/18 14:14:33 jn Exp $
*/

#ifndef EXT2FS_H
#define EXT2FS_H

#include <gplntifs.h>
#include <ntdddisk.h>

#ifndef _GNU_NTIFS_
#define PEXTENDED_IO_STACK_LOCATION PIO_STACK_LOCATION
#endif


#define DRIVER_NAME "[EXT2]"
#define DEVICE_NAME L"\\ext2"

#define FSD_REGISTRY_PATH L"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\Ext2fs"

#define EXT2_SUPER_MAGIC		0xEF53
#define EXT2_SUPER_MAGIC_OFFSET	56
#define EXT2_SUPER_OFFSET		1024


/**************************************************************************
	some useful definitions
**************************************************************************/
#ifdef	_CPU_X86_
typedef	char				int8;
typedef	short				int16;
typedef	long				int32;

typedef	unsigned char		uint8;
typedef	unsigned short		uint16;
typedef	unsigned long		uint32;

typedef unsigned char		__u8;
typedef unsigned short		__u16;
typedef unsigned long		__u32;

typedef char	__s8;
typedef short	__s16;
typedef long	__s32;

#include <linux\ext2_fs.h>

typedef struct ext2_super_block	SUPER_BLOCK, *PSUPER_BLOCK;
typedef struct ext2_group_desc GROUP_DESC, *PGROUP_DESC;
typedef struct ext2_inode INODE, *PINODE;
typedef void *PBITMAP;

typedef PVOID PBCB;

#define S_IFMT  00170000
#define S_IFSOCK 0140000
#define S_IFLNK	 0120000
#define S_IFREG  0100000
#define S_IFBLK  0060000
#define S_IFDIR  0040000
#define S_IFCHR  0020000
#define S_IFIFO  0010000
#define S_ISUID  0004000
#define S_ISGID  0002000
#define S_ISVTX  0001000

#define S_ISLNK(m)	(((m) & S_IFMT) == S_IFLNK)
#define S_ISREG(m)	(((m) & S_IFMT) == S_IFREG)
#define S_ISDIR(m)	(((m) & S_IFMT) == S_IFDIR)
#define S_ISCHR(m)	(((m) & S_IFMT) == S_IFCHR)
#define S_ISBLK(m)	(((m) & S_IFMT) == S_IFBLK)
#define S_ISFIFO(m)	(((m) & S_IFMT) == S_IFIFO)
#define S_ISSOCK(m)	(((m) & S_IFMT) == S_IFSOCK)

#define S_IRWXU 00700
#define S_IRUSR 00400
#define S_IWUSR 00200
#define S_IXUSR 00100

#define S_IRWXG 00070
#define S_IRGRP 00040
#define S_IWGRP 00020
#define S_IXGRP 00010

#define S_IRWXO 00007
#define S_IROTH 00004
#define S_IWOTH 00002
#define S_IXOTH 00001

// we will use the LARGE_INTEGER structure as defined by NT

#else		// Please define appropriate types here

!!!! You must define your types here for compilation to proceed !!!!

#endif	// if _CPU_X86_


// Resource locks
#define ACQUIRE(s) ExAcquireResourceSharedLite(&(s), TRUE)
#define ACQUIRE_EX(s) ExAcquireResourceExclusiveLite(&(s), TRUE)
#define RELEASE(s) ExReleaseResourceForThreadLite(&(s), ExGetCurrentResourceThread())

#define try_return(s) { (s); goto try_exit; }
//
//  This macro returns TRUE if a flag in a set of flags is on and FALSE
//  otherwise.  It is followed by two macros for setting and clearing
//  flags
//

#define BooleanFlagOn(Flags,SingleFlag) ((BOOLEAN)((((Flags) & (SingleFlag)) != 0)))

#define SetFlag(Flags,SingleFlag) { \
    (Flags) |= (SingleFlag);        \
}


#define ClearFlag(Flags,SingleFlag) { \
    (Flags) &= ~(SingleFlag);         \
}

///////////////////////////////
// Data structures
///////////////////////////////

//
// FSD_IDENTIFIER_TYPE
//
// Identifiers used to mark the structures
//
typedef enum _FSD_IDENTIFIER_TYPE {
    FGD = ':DGF', // Global Data
    VCB = ':VCB', // Volume Control Block
    FCB = ':FCB', // File Control Block
    CCB = ':CCB', // Context Control Block
#ifndef FSD_RO
    FLI = ':ILF',
#endif // !FSD_RO
    ICX = ':XCI',
	BGB = ':BGB', // Block Group
	BLK = ':BLK',
	INO = ':INO',
	DIR = ':DIR'
} FSD_IDENTIFIER_TYPE;

//
// FSD_IDENTIFIER
//
// Header put in the beginning of every structure
//
typedef struct _FSD_IDENTIFIER {
    FSD_IDENTIFIER_TYPE     Type;
    ULONG                   Size;
} FSD_IDENTIFIER, *PFSD_IDENTIFIER;


#define VERIFY(s, t) ASSERT(((PFSD_IDENTIFIER)s)->Type == t)


typedef struct _XtFSGlobals {
	FSD_IDENTIFIER				Identifier;
	// the fields in this list are protected by the following resource
	ERESOURCE					Resource;
	// each driver has a driver object created for it by the NT I/O Mgr.
	//	we are no exception to this rule.
	PDRIVER_OBJECT				DriverObject;
	// we will create a device object for our FSD as well ...
	//	Although not really required, it helps if a helper application
	//	writen by us wishes to send us control information via
	//	IOCTL requests ...
	PDEVICE_OBJECT				DeviceObject;
	// we will keep a list of all logical volumes for our sample FSD
	LIST_ENTRY					NextVCB;

    //
    //  Cache manager call back structures, which must be passed on each call
    //  to CcInitializeCacheMap.
    //

    CACHE_MANAGER_CALLBACKS CacheManagerCallbacks;
    CACHE_MANAGER_CALLBACKS CacheManagerNoOpCallbacks;


} XtFSGlobals, *PtrXtFSGlobals;

extern XtFSGlobals *Global;


//
typedef struct _FILE_ITERATOR {
	__u32	BlockNumber;
	__u32	IBlockNumber;
	__u32	DBlockNumber;
	__u32	TBlockNumber;
} XtFSFILEITERATOR, *PFILEITERATOR;
//


typedef struct _BLOCK_OBJECT {
    FSD_IDENTIFIER      Identifier;
	LIST_ENTRY			NextObject;
	__u32				BlockNo;
	__u32				BlockCount;
	__u32				RefCount;
	PBCB				Bcb;
	PVOID				Buffer;

} XtBGBLOCK_OBJECT, *PBLOCK_OBJECT;

// this is what you get to work with
// INodes.  The inode it's self is allocated
// and maintained by the block group.
typedef struct _INODE_OBJECT {
    FSD_IDENTIFIER      Identifier;
	LIST_ENTRY			NextObject;
	__u32				INodeNo;
	__u32				RefCount;
	PBLOCK_OBJECT		Block;
	PINODE				INode;

} XtBGINODE_OBJECT, *PINODE_OBJECT;



// used for enumerating a directory inode
typedef struct _DIR_OBJECT {
    FSD_IDENTIFIER      Identifier;
	PINODE_OBJECT		INode;

	__u32		  BlockSize;
	__u32		  BlockNumber;
	__u32		  Offset;
	PBLOCK_OBJECT Block;
	PVOID		  Vcb;
	PCHAR		  Filter;
	ULONG		  FilterLen;	

} XtBGDIR_OBJECT, *PDIR_OBJECT;

typedef struct _BLOCK_GROUP {

    // Identifier for this structure
    FSD_IDENTIFIER      Identifier;
	__u32				GroupNo;
	__u32				RefCount;
	PDEVICE_OBJECT		DeviceObject;
	PBITMAP				BlockBitmap;
	PBCB				BlockBitmapBcb;
	PBITMAP				INodeBitmap;
	PBCB				INodeBitmapBcb;

	PGROUP_DESC			Info;
	__u32				StartBlock;


} XtBGBLOCK_GROUP, *PBLOCK_GROUP;


typedef struct _XtFSVolumeControlBlock {
	FSD_IDENTIFIER				Identifier;
    // FCB header required by NT
    // (the VCB is also used as an FCB for the file
    //  object that represents the volume itself)
	// this in the object which must be used as the FsContext
	// to get a CVB from it use a CONTAINING_RECORD
    FSRTL_COMMON_FCB_HEADER     CommonFCBHeader;

    //
    //  The following field points to the file object used to do I/O to
    //  the virtual volume file.  The virtual volume file maps sectors
    //  0 through the end of fat and is of a fixed size (determined during
    //  mount)
    //

    PFILE_OBJECT VirtualVolumeFile;

    // Information about the physical device object
    DISK_GEOMETRY               DiskGeometry;
    PARTITION_INFORMATION       PartitionInformation;

    //
    //  The following field contains a record of special pointers used by
    //  MM and Cache to manipluate section objects.  Note that the values
    //  are set outside of the file system.  However the file system on an
    //  open/create will set the file object's SectionObject field to point
    //  to this field
    //

    SECTION_OBJECT_POINTERS SectionObjectPointers;

	// These are used by the FsContext
	ERESOURCE	MainResource;
	ERESOURCE	PagingIoResource;



	// a resource to protect the fields contained within the VCB
	ERESOURCE							Resource;
	// each VCB is accessible off a global linked list
	LIST_ENTRY							NextVCB;
	// each VCB points to a VPB structure created by the NT I/O Manager
	PVPB									PtrVPB;

	// these are needed to support the disk structure access...
	PSUPER_BLOCK	SuperBlock;
	__u32			BLBlockSize;
	PBLOCK_GROUP	*BGBlockGroups;
	__u32			BGGroupCount;
	PGROUP_DESC		BGGroupDescriptors;
	PBCB			BGGroupDescriptorsBcb;
	__u32			BGDescriptorSize;
	__u32			BGBlockMapSize;
	__u32			BGINodeMapSize;
	__u32			BGTableSize;
	__u32			BGBlockHeadSize;

	__u32			INodesPerBlock;

	__u32			BLIndMax;  // max block number in the indirect block
	__u32			BLDIndMax; // max block number in the double indirect block
	__u32			BLTIndMax; // max block number in the tripple indirect block
	///


	// a set of flags that might mean something useful
	//uint32								VCBFlags;
	// A count of the number of open files/directories
	//	As long as the count is != 0, the volume cannot
	//	be dismounted or locked.
	//uint32								VCBOpenCount;
	// a global list of all FCB structures associated with the VCB
	LIST_ENTRY							NextFCB;
	// we will maintain a global list of IRP's that are pending
	//	because of a directory notify request.
	LIST_ENTRY							NextNotifyIRP;
	// the above list is protected only by the mutex declared below
	KMUTEX								NotifyIRPMutex;
	// for each mounted volume, we create a device object. Here then
	//	is a back pointer to that device object
	PDEVICE_OBJECT						VCBDeviceObject;
	// We also retain a pointer to the physical device object on which we
	// have mounted ourselves. The I/O Manager passes us a pointer to this
	// device object when requesting a mount operation.
	PDEVICE_OBJECT						TargetDeviceObject;
	// the volume structure contains a pointer to the root directory FCB
//	PtrSFsdFCB							PtrRootDirectoryFCB;
	// the complete name of the user visible drive letter we serve
	uint8									*PtrVolumePath;
	// For volume open operations, we do not create a FCB (we use the VCB
	//	directly instead). Therefore, all CCB structures for the volume
	//	open operation are linked directly off the VCB
	LIST_ENTRY							VolumeOpenListHead;
	// Pointer to a stream file object created for the volume information
	// to be more easily read from secondary storage (with the support of
	// the NT Cache Manager).
	PFILE_OBJECT						PtrStreamFileObject;
	// Required to use the Cache Manager.
   SECTION_OBJECT_POINTERS			SectionObject;
	// File sizes required to use the Cache Manager.
	LARGE_INTEGER						AllocationSize;
	LARGE_INTEGER						FileSize;
	LARGE_INTEGER						ValidDataLength;

	ERESOURCE						BlockResource;
	LIST_ENTRY						NextBlockObject;
	ERESOURCE						INodeResource;
	LIST_ENTRY						NextINodeObject;

} XtFSVCB, *PtrXtFSVCB, *PVCB;

// some valid flags for the VCB
#define			SFSD_VCB_FLAGS_VOLUME_MOUNTED			(0x00000001)
#define			SFSD_VCB_FLAGS_VOLUME_LOCKED			(0x00000002)
#define			SFSD_VCB_FLAGS_BEING_DISMOUNTED		(0x00000004)
#define			SFSD_VCB_FLAGS_SHUTDOWN					(0x00000008)
#define			SFSD_VCB_FLAGS_VOLUME_READ_ONLY		(0x00000010)

#define			SFSD_VCB_FLAGS_VCB_INITIALIZED		(0x00000020)



//
// FSD_IRP_CONTEXT
//
// Used to pass information about a request between the drivers functions
//
typedef struct _FSD_IRP_CONTEXT {

    // Identifier for this structure
    FSD_IDENTIFIER      Identifier;


	// can we save a few entries by putting out IoStack pointer in here?

    // Pointer to the IRP this request describes
    PIRP                Irp;

    // The major and minor function code for the request
    UCHAR               MajorFunction;
    UCHAR               MinorFunction;

    // The device object
    PDEVICE_OBJECT      DeviceObject;

    // The file object
    PFILE_OBJECT        FileObject;

    // If the request is synchronous (we are allowed to block)
    BOOLEAN             IsSynchronous;

    // If the request is top level
    BOOLEAN             IsTopLevel;

    // Used if the request needs to be queued for later processing
    WORK_QUEUE_ITEM     WorkQueueItem;

    // If an exception is currently in progress
    BOOLEAN             ExceptionInProgress;

    // A possible exception code
    NTSTATUS            ExceptionCode;

	BOOLEAN				Completed;
	BOOLEAN				FreeContext;
	BOOLEAN				Pending;

} FSD_IRP_CONTEXT, *PFSD_IRP_CONTEXT;

typedef struct _FSD_ALLOC_HEADER {
    FSD_IDENTIFIER Identifier;
} FSD_ALLOC_HEADER, *PFSD_ALLOC_HEADER;



// one FCB for each file (inode)
// no matter how many times the file is opened
typedef struct _XtFSFileControlBlock
{
	FSD_IDENTIFIER				Identifier;
    // FCB header required by NT
	// this in the object which must be used as the FsContext
	// to get a CVB from it use a CONTAINING_RECORD
    FSRTL_COMMON_FCB_HEADER     CommonFCBHeader;

	// inode pointer ?
	__u32 inode_number;
	PINODE_OBJECT INode;

	// pointer to out vcb (block device)
	PtrXtFSVCB	vcb;

    //
    //  The following field contains a record of special pointers used by
    //  MM and Cache to manipluate section objects.  Note that the values
    //  are set outside of the file system.  However the file system on an
    //  open/create will set the file object's SectionObject field to point
    //  to this field
    //

    SECTION_OBJECT_POINTERS SectionObjectPointers;

	// These are used by the FsContext
	ERESOURCE	MainResource;
	ERESOURCE	PagingIoResource;

	// ??
	LIST_ENTRY	NextFCB;
	ERESOURCE	Resource;

	// a list of CCB's 
	LIST_ENTRY	NextCCB;

	uint32 ReferenceCount;
	uint32 OpenHandleCount;

	// what else does an FCB require?
} XtFSFCB, *PtrXtFSFCB, *PFCB;


// one CCB for each open file handle (vnode)
// These all point to the FCB for that inode
typedef struct _XtFSContextControlBlock
{
	FSD_IDENTIFIER		Identifier;
	ERESOURCE			Resource;

	PtrXtFSFCB fcb;

	// a list of CCB's 
	LIST_ENTRY	NextCCB;

	// this is here and not in the fcb because the same INode can have
	// different names from hard links.
	PCHAR		  FileName;
	ULONG		  FileNameLen;	


	PDIR_OBJECT DirEnumObject; // if this assigned then a query is in progress

	int file_count;

	// these fields are for file opens

	//

} XtFSCCB, *PtrXtFSCCB, *PCCB;


#if DBGasdf


#define DbgIoCompleteRequest(Irp, PriorityBoost) \
        FsdDbgPrintComplete(Irp, PriorityBoost); \
        IoCompleteRequest(Irp, PriorityBoost)

#else

#define DbgIoCompleteRequest(Irp, PriorityBoost) \
        IoCompleteRequest(Irp, PriorityBoost)

#endif


extern LARGE_INTEGER LargeIntegerZero;
extern LARGE_INTEGER LargeIntegerMax;



////////////////////
// ext2fs.c
////////////////////

NTSTATUS
DriverEntry (
	IN PDRIVER_OBJECT	DriverObject,
	IN PUNICODE_STRING	RegistryPath
    );

VOID
XtFSDriverUnload (
	IN PDRIVER_OBJECT	DriverObject
    );

NTSTATUS
XtFSBuildRequest (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    );

NTSTATUS
XtFSQueueRequest (
    IN PFSD_IRP_CONTEXT IrpContext
    );

VOID
XtFSDeQueueRequest (
    IN PVOID Context
    );

NTSTATUS
XtFSDispatchRequest (
    IN PFSD_IRP_CONTEXT IrpContext
    );

NTSTATUS
XtFSExceptionFilter (
    IN PFSD_IRP_CONTEXT     IrpContext,
    IN PEXCEPTION_POINTERS  ExceptionPointers
    );

NTSTATUS
XtFSExceptionHandler (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PIRP             Irp
    );

NTSTATUS
XtFSLockUserBuffer (
    IN PIRP             Irp,
    IN ULONG            Length,
    IN LOCK_OPERATION   Operation
    );

PVOID
XtFSGetUserBuffer (
    IN PIRP Irp
    );

PFSD_IRP_CONTEXT
XtFSAllocateIrpContext (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    );

	VOID
XtFSFreeIrpContext (
    IN PFSD_IRP_CONTEXT IrpContext
    );

VOID 
XtFSCompleteRequest(
		IN PFSD_IRP_CONTEXT  IrpContext,
		IN CCHAR  PriorityBoost
		);

////////////////////
// fsctl.c
////////////////////

NTSTATUS
XtFSFileSystemControl (
    IN PFSD_IRP_CONTEXT IrpContext
    );




////////////////////
// create.c
////////////////////

// this is a debug function
NTSTATUS UnicodePrint(PCH Msg, PUNICODE_STRING String);

PtrXtFSFCB XtFSFindFCBbyINode(PtrXtFSVCB vcb, __u32 inode);
PtrXtFSFCB XtFSAllocateFCB(PtrXtFSVCB vcb, __u32 inode);
NTSTATUS XtFSFreeFCB(PtrXtFSFCB fcb);
PtrXtFSCCB XtFSAllocateCCB(PtrXtFSFCB fcb);
NTSTATUS XtFSFreeCCB(PtrXtFSCCB ccb);


NTSTATUS
XtFSCreate (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// close.c
////////////////////

NTSTATUS
XtFSClose (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// read.c
////////////////////

NTSTATUS
XtFSRead (
    IN PFSD_IRP_CONTEXT IrpContext
	);

NTSTATUS
XtBLReadBlockDeviceSync (
    IN PFSD_IRP_CONTEXT IrpContext,
	IN PtrXtFSVCB		Vcb,
    IN PLARGE_INTEGER   Offset,
    IN ULONG            Length,
    IN OUT PVOID        Buffer
    );

////////////////////
// write.c
////////////////////

NTSTATUS
XtFSWrite (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// cleanup.c
////////////////////

NTSTATUS
XtFSCleanup (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// dirctl.c
////////////////////

NTSTATUS
XtFSDirectoryControl (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// volinfo.c
////////////////////

NTSTATUS
XtFSQueryVolumeInformation (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// fileinfo.c
////////////////////

NTSTATUS
XtFSQueryInformation (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// devctl.c
////////////////////

NTSTATUS
XtDeviceControl (
    IN PFSD_IRP_CONTEXT IrpContext
	);

////////////////////
// superblock.c
////////////////////


NTSTATUS
XtSBCheckDeviceForExt2fs (
	IN PDEVICE_OBJECT	DeviceObject
);

NTSTATUS 
XtSBMountFS(
	PtrXtFSVCB Vcb
);


////////////////////
// block.c
////////////////////

NTSTATUS
XtBLReadBlockDevice (
	IN PDEVICE_OBJECT	DeviceObject,
	IN PLARGE_INTEGER	Offset,
	IN ULONG			Length,
	OUT PVOID			Buffer
);

NTSTATUS 
XtBLInitBlockIO(
	PtrXtFSVCB Vcb
);

__u32 XtDiv2(
	__u32 Numerator,
	__u32 Denominator
);

__u32 XtDiv1(
	__u32 Neumerator,
	__u32 Denominator,
	__u32 *Remainder
);

PVOID
XtBLGetBlocks(
	PtrXtFSVCB Vcb,
	__u32 BlockNo,
	__u32 BlockCount,
	PBCB *BlockBcb
);

NTSTATUS
XtBLFreeBlocks(
	PtrXtFSVCB Vcb,
	PBCB *BlockBcb
);

NTSTATUS 
XtBGGetBlockGroup(
	PtrXtFSVCB Vcb,
	__u32 GroupNumber,
	PBLOCK_GROUP *Pointer
);

NTSTATUS 
XtBGFreeBlockGroup(
	PtrXtFSVCB Vcb,
	__u32 BlockNumber,
	PBLOCK_GROUP *Pointer
);

NTSTATUS
XtLockBlock(
	PtrXtFSVCB Vcb,
	__u32 BlockNumber,
	__u32 BlockCount,
	PBLOCK_OBJECT *Pointer
);

NTSTATUS
XtMakeBlockPersistant(
	PtrXtFSVCB Vcb,
	PBLOCK_OBJECT *Pointer
);

NTSTATUS
XtUnLockBlock(
	PtrXtFSVCB Vcb,
	PBLOCK_OBJECT *Pointer
);

NTSTATUS
XtLockINode(
	PtrXtFSVCB Vcb,
	__u32 INodeNumber,
	PINODE_OBJECT *Pointer
);

NTSTATUS
XtUnLockINode(
	PtrXtFSVCB Vcb,
	PINODE_OBJECT *Pointer
);

NTSTATUS
XtMakeINodePersistant(
	PtrXtFSVCB Vcb,
	PINODE_OBJECT *Pointer
);

NTSTATUS
XtCreateDirObject(
	PtrXtFSVCB Vcb,
	PINODE_OBJECT INode,
	PDIR_OBJECT *Dir,
	PSTRING Filter
);

NTSTATUS
XtMakeDirObjectPersistant(
	PDIR_OBJECT *Dir
);

NTSTATUS
XtDestroyDirObject(
	PDIR_OBJECT *Dir
);

NTSTATUS
XtDirGetEntry(
	PDIR_OBJECT Dir,
	WCHAR *name_buffer,
	ULONG *name_len,
	__u32 *inode
);

NTSTATUS
XtDirFindEntry(
	PDIR_OBJECT Dir,
	PCHAR name_buffer,
	ULONG name_len,
	__u32 *inode
);

NTSTATUS
XtBLReadFile (
    IN PFSD_IRP_CONTEXT IrpContext,
	IN PtrXtFSVCB		Vcb,
	IN PtrXtFSCCB		Ccb,
    IN PLARGE_INTEGER   Offset,
    IN ULONG            Length,
    IN OUT PCHAR        Buffer
    );

void XtUnixTimeToNTTime(__u32 *UnixTime, PLARGE_INTEGER NTTime);
void XtNTTimeToUnixTime(PLARGE_INTEGER NTTime, __u32 *UnixTime);



////////////////////
// mem.c
////////////////////

PVOID
XtMMAllocatePool (
    IN POOL_TYPE    PoolType,
    IN ULONG        NumberOfBytes
    );

PVOID
XtMMAllocatePoolStruct (
    IN POOL_TYPE    PoolType,
    IN ULONG        NumberOfBytes,
	IN FSD_IDENTIFIER_TYPE StructType
    );

VOID
XtMMFreePool (
    IN PVOID P,
    IN ULONG NumberOfBytes
    );

/////////////////////
// resource.c
/////////////////////
VOID
XtFSNoOpRelease (
    IN PVOID Fcb
    );

BOOLEAN
XtFSNoOpAcquire (
    IN PVOID Fcb,
    IN BOOLEAN Wait
    );


//VOID HalDisplayString(PSZ Buffer);


#endif
