/*
	$Header: /home/itig/cvsroot/ext2ifs/src/create.c,v 1.17 2001/10/17 13:53:24 jn Exp $
*/

#include "ext2fs.h"

CHAR *File_Disposition[] =
{
	"FILE_SUPERSEDE",
	"FILE_OPEN",
	"FILE_CREATE",
	"FILE_OPEN_IF",
	"FILE_OVERWRITE",
	"FILE_OVERWRITE_IF"
};


NTSTATUS UnicodePrint(PCH Msg, PUNICODE_STRING String)
{
	int i;
	char str[512];

	KdPrint((Msg));

	
	if(String)
	{
		for(i=0; i<(String->Length / 2); i++)
		{
			if(i >= 512)
			{
				break; // string too long for now...
			}
			str[i] = (char)String->Buffer[i];
			//KdPrint(("%c", String->Buffer[i]));
		}
		str[i++] = 0;
		KdPrint(("%s\n", str));
	}
	else
	{
		KdPrint(("NULL\n"));
	}
	return STATUS_SUCCESS;
}


PtrXtFSFCB XtFSFindFCBbyINode(PtrXtFSVCB vcb, __u32 inode)
{
	PtrXtFSFCB fcb = 0;
	void *start;
	int item_count = 0;

	// walk through the FCB's in the VCB
	// if we find a match then increment all the counters
	// and return the pointer
	start = vcb->NextFCB.Flink;

	KdPrint(("[EXT2]: Looking for FCB for inode %d\n", inode));

	while(start != &vcb->NextFCB)
	{
		item_count++;

		// look at start...
		fcb = CONTAINING_RECORD(start, XtFSFCB, NextFCB);
		//KdPrint(("[EXT2]: Item %d inode = %d\n", item_count, fcb->inode));
		// make sure that fcb is an fcb....?

		// see if it matches...?
		if(fcb->inode_number == inode)
		{
			KdPrint(("[EXT2]: Existing FCB found\n"));
			return fcb;
		}

		start = fcb->NextFCB.Flink;
	}
	KdPrint(("[EXT2]: Existing FCB not found\n"));

	return 0;
}

PtrXtFSFCB XtFSAllocateFCB(PtrXtFSVCB vcb, __u32 inode)
{
	PtrXtFSFCB new_fcb;

	//create the FCB
    //new_fcb = ExAllocatePool(NonPagedPool, sizeof(XtFSFCB));
    //RtlZeroMemory(new_fcb, sizeof(XtFSFCB));
	new_fcb = XtMMAllocatePoolStruct(NonPagedPool, sizeof(XtFSFCB), FCB);

	// set pointer to VCB
    new_fcb->vcb = vcb;//((PtrXtFSVCB)IrpContext->DeviceObject->DeviceExtension);

    ExInitializeResourceLite(&(new_fcb->MainResource));
    ExInitializeResourceLite(&(new_fcb->PagingIoResource));

    new_fcb->CommonFCBHeader.Resource = &(new_fcb->MainResource);
    new_fcb->CommonFCBHeader.PagingIoResource = &(new_fcb->PagingIoResource);

    new_fcb->CommonFCBHeader.AllocationSize.QuadPart = 4096;
    new_fcb->CommonFCBHeader.FileSize.QuadPart = 4096;

    new_fcb->CommonFCBHeader.ValidDataLength.LowPart = 0xffffffff;
    new_fcb->CommonFCBHeader.ValidDataLength.HighPart = 0x7fffffff;

    new_fcb->CommonFCBHeader.IsFastIoPossible = 0;//FastIoIsPossible;

	// link into the vcb
	new_fcb->inode_number = inode;
	KdPrint(("[EXT2]: Created inode %d\n", inode));

	ExInitializeResourceLite(&new_fcb->Resource);
    InitializeListHead(&new_fcb->NextCCB);


	// map in the inode....
	XtLockINode(vcb, inode, &(new_fcb->INode));

	// what block do we need?

	InsertTailList(&(vcb->NextFCB), &(new_fcb->NextFCB));

	return new_fcb;
}

NTSTATUS XtFSFreeFCB(PtrXtFSFCB fcb)
{
	RemoveEntryList(&(fcb->NextFCB));
	XtUnLockINode(fcb->vcb, &(fcb->INode));
	ExDeleteResourceLite(&fcb->Resource);
	ExDeleteResourceLite(&fcb->MainResource);
	ExDeleteResourceLite(&fcb->PagingIoResource);
	XtMMFreePool(fcb, sizeof(XtFSFCB));

	return STATUS_SUCCESS;
}


PtrXtFSCCB XtFSAllocateCCB(PtrXtFSFCB fcb)
{
	PtrXtFSCCB ccb = XtMMAllocatePoolStruct(NonPagedPool, sizeof(XtFSCCB), CCB);
	ASSERT(ccb);
	VERIFY(ccb, CCB);
	ccb->fcb = fcb;
    
	ExInitializeResourceLite(&ccb->Resource);

//	KdPrint(("[EXT2]: Linking ccb into fcb's list\n"));

	// link ccb into the fcb list of ccb's
	InsertTailList(&(fcb->NextCCB), &(ccb->NextCCB));

	return ccb;
}

NTSTATUS XtFSFreeCCB(PtrXtFSCCB ccb)
{
	// unlink from the fcb
	RemoveEntryList(&(ccb->NextCCB));

	// free the memory
	ASSERT(ccb->Resource);
	ExDeleteResourceLite(&ccb->Resource);

	XtMMFreePool(ccb, sizeof(XtFSCCB));

	return STATUS_SUCCESS;
}

void PrintSegment(char *Segment, int SegmentLen)
{
	char buffer[512];
	int i;

	KdPrint(("[EXT2]: Segment len = %d: ", SegmentLen));
	for(i = 0; i < SegmentLen; i++)
	{
		if(i >= 512)
		{
			break;
		}
		buffer[i] = Segment[i];
		//KdPrint(("%c", Segment[i]));
	}
	buffer[i++] = 0;
	KdPrint(("%s\n", buffer));
}

NTSTATUS
XtFSCreate (
    IN PFSD_IRP_CONTEXT IrpContext
	)
{
    IN PIRP             Irp = IrpContext->Irp;
	NTSTATUS			Status = STATUS_SUCCESS;
	ULONG				Info = 0;
	ULONG				Options;
	PIO_STACK_LOCATION		PtrIoStackLocation;
	PFILE_OBJECT			PtrNewFileObject = NULL;
	PFILE_OBJECT			PtrRelatedFileObject = NULL;
	PUNICODE_STRING			TargetObjectName;
	PUNICODE_STRING			RelatedObjectName;
	int i;
	int Disposition;
	PtrXtFSVCB vcb;

    PtrXtFSFCB            root_fcb;
    PtrXtFSCCB            root_ccb;


	KdPrint(("[EXT2]: Create request\n"));

	PtrIoStackLocation = IoGetCurrentIrpStackLocation(Irp);

	ASSERT(PtrIoStackLocation);

	// the flags...
	KdPrint(("[EXT2]: Create Options        0x%08x\n", PtrIoStackLocation->Parameters.Create.Options & FILE_VALID_OPTION_FLAGS));
	Options = PtrIoStackLocation->Parameters.Create.Options & FILE_VALID_OPTION_FLAGS;
//	KdPrint(("[EXT2]: Create Options        %s\n", File_Disposition[PtrIoStackLocation->Parameters.Create.Options]));
	Disposition = (PtrIoStackLocation->Parameters.Create.Options >> 24) & 0xFF;
	if(Disposition <= FILE_MAXIMUM_DISPOSITION)
	{
		KdPrint(("[EXT2]: Create Options disposition (%d) %s\n", Disposition, File_Disposition[Disposition] ));
	}
	else
	{
		KdPrint(("[EXT2]: Create Options disposition %d is too big\n", Disposition));
	}



	KdPrint(("[EXT2]: Create FileAttributes 0x%04x\n", PtrIoStackLocation->Parameters.Create.FileAttributes));


	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_READONLY)
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_READONLY\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_HIDDEN)
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_HIDDEN\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_SYSTEM) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_SYSTEM\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_DIRECTORY) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_DIRECTORY\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_ARCHIVE) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_ARCHIVE\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_NORMAL) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_NORMAL\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_TEMPORARY) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_TEMPORARY\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_COMPRESSED) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_COMPRESSED\n"));

	if(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_OFFLINE) 
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_OFFLINE\n"));

/*	(PtrIoStackLocation->Parameters.Create.FileAttributes & FILE_ATTRIBUTE_PROPERTY_SET) ?
		KdPrint(("[EXT2]: Create FileAttributes FILE_ATTRIBUTE_PROPERTY_SET\n")):0;*/ // This is not in the MS ifs.h



	KdPrint(("[EXT2]: Create ShareAccess    0x%04x\n", PtrIoStackLocation->Parameters.Create.ShareAccess));

	(PtrIoStackLocation->Parameters.Create.ShareAccess & FILE_SHARE_READ) ?
		KdPrint(("[EXT2]: Create ShareAccess FILE_SHARE_READ\n")):0;

	(PtrIoStackLocation->Parameters.Create.ShareAccess & FILE_SHARE_WRITE) ?
		KdPrint(("[EXT2]: Create ShareAccess FILE_SHARE_WRITE\n")):0;

	(PtrIoStackLocation->Parameters.Create.ShareAccess & FILE_SHARE_DELETE) ?
		KdPrint(("[EXT2]: Create ShareAccess FILE_SHARE_DELETE\n")):0;

	KdPrint(("[EXT2]: Create EaLength       0x%08x\n", PtrIoStackLocation->Parameters.Create.EaLength));
	


	PtrNewFileObject	= PtrIoStackLocation->FileObject;
	ASSERT(PtrNewFileObject);

    TargetObjectName	= &PtrNewFileObject->FileName;

	if(TargetObjectName->Length == 0)
	{
		// we are opening the device
		KdPrint(("[EXT2]: Create request for FS device\n"));
		vcb = (PtrXtFSVCB)IrpContext->DeviceObject->DeviceExtension;
		//PtrIoStackLocation->FileObject->FsContext = &(vcb->CommonFCBHeader);

		Status = STATUS_SUCCESS;
	}
	else
	{
		// we are opening a file (or directory)
		PtrRelatedFileObject = PtrNewFileObject->RelatedFileObject;

		UnicodePrint("[EXT2]: Create request for ", TargetObjectName);
		/*if(PtrRelatedFileObject)
		{
			UnicodePrint("[EXT2]: Related file object ", &(PtrRelatedFileObject->FileName));
		}
		else
		{
			KdPrint(("[EXT2]: No related object\n"));
		}
		KdPrint(("[EXT2]: Name Length = %d\n", TargetObjectName->Length));*/

		vcb = (PtrXtFSVCB)IrpContext->DeviceObject->DeviceExtension;

		// open the root directory...
		if((PtrRelatedFileObject == NULL) && (TargetObjectName->Length == 2))
		{
			KdPrint(("[EXT2]: Create request for root directory\n"));

			// The rood inode is always EXT2_ROOT_INO
			root_fcb = XtFSFindFCBbyINode(vcb, EXT2_ROOT_INO);

			if(root_fcb == 0)
			{
				// we need to allocate one
				root_fcb = XtFSAllocateFCB(vcb, EXT2_ROOT_INO);
			}

			root_ccb = XtFSAllocateCCB(root_fcb);
            
            PtrIoStackLocation->FileObject->FsContext = &(root_fcb->CommonFCBHeader);
            PtrIoStackLocation->FileObject->FsContext2 = root_ccb;
            PtrIoStackLocation->FileObject->PrivateCacheMap = NULL;
            PtrIoStackLocation->FileObject->SectionObjectPointer = &(root_fcb->SectionObjectPointers);

            Irp->IoStatus.Information = FILE_OPENED;
            Status = STATUS_SUCCESS;
            KdPrint((DRIVER_NAME ": Create on \\ succeded\n"));
		}
		else
		{
			ULONG Length = 0;
			PCHAR Target = NULL;
			ULONG RelLen = 0;
			ULONG TargetLen = 0;
			ULONG ExtraLen = 0;
			PINODE_OBJECT INode = NULL;
			PDIR_OBJECT Directory = NULL;
			__u32 NextINode;
			PCHAR Segment = 0;
			ULONG SegmentLen = 0;
			
			ULONG i = 0;
			// we need to look up the file.....

			// first, is there a starting point...
			if(PtrRelatedFileObject)
			{
	            KdPrint((DRIVER_NAME ": Using related file object\n"));
				RelLen = PtrRelatedFileObject->FileName.Length / 2;
				if(PtrRelatedFileObject->FileName.Buffer[RelLen - 1] != '\\')
				{
					ExtraLen = 1;
				}
			}
			TargetLen = TargetObjectName->Length / 2;
            KdPrint((DRIVER_NAME ": Target len = %d\n", TargetLen));
			Length = RelLen + ExtraLen + TargetLen;

			Target = ExAllocatePool(NonPagedPool, Length);

			if(PtrRelatedFileObject)
			{
				for(i = 0; i < RelLen; i++)
				{
					Target[i] = (char)PtrRelatedFileObject->FileName.Buffer[i];
				}
			}

			// sometimes we need and extra slash
			if(ExtraLen > 0)
			{
				Target[RelLen] = '\\';
			}

			// make one ansi char string
			for(i = 0; i < TargetLen; i++)
			{
				Target[i + RelLen + ExtraLen] = (char)TargetObjectName->Buffer[i];
			}
			PrintSegment(Target, Length);
			if(Target[Length - 1] == '\\')
			{
				KdPrint(("[EXT2]: Warning: Target ends with a \\\n"));
				Length--;
			}


			// we now have an absolute path name to the file...
			NextINode = EXT2_ROOT_INO;

			for(i = 0; i < Length; i++)
			{
				if(Target[i] == '\\')
				{
					// process the segment
					if(SegmentLen > 0)
					{
						PrintSegment(Segment, SegmentLen);

						XtLockINode(vcb, NextINode, &INode);
						// check that the inode is a directory
						XtCreateDirObject(vcb, INode, &Directory, NULL);
						Status = XtDirFindEntry(Directory, Segment, SegmentLen, &NextINode);
						XtDestroyDirObject(&Directory);
						XtUnLockINode(vcb, &INode);
						if(Status == STATUS_SUCCESS)
						{
							// we have a match...
							KdPrint(("[EXT2]: Found %d\n", NextINode));
						}
						else
						{
							break;
						}
					}
					SegmentLen = 0;
					Segment = Target + i + 1; 
				}
				else
				{
					SegmentLen++;
				}
			}
			if((SegmentLen > 0) && (Status == STATUS_SUCCESS))
			{
				// there is a file name
				PrintSegment(Segment, SegmentLen);
				XtLockINode(vcb, NextINode, &INode);
				// check that the inode is the correct type a directory
				XtCreateDirObject(vcb, INode, &Directory, NULL);
				Status = XtDirFindEntry(Directory, Segment, SegmentLen, &NextINode);
				XtDestroyDirObject(&Directory);
				XtUnLockINode(vcb, &INode);
				if(Status == STATUS_SUCCESS)
				{
					// we have a match...
					KdPrint(("[EXT2]: Found %d\n", NextINode));

					// open the INode...
					XtLockINode(vcb, NextINode, &INode);
					// check that we still want to open it... ie is a file not a dir

					// see if we need a directory...
					if(Options & FILE_DIRECTORY_FILE)
					{
						KdPrint(("[EXT2]: We must open a directory and flags are 0x%08X\n", INode->INode->i_mode & S_IFMT));
						if(S_ISDIR(INode->INode->i_mode))
						{
							KdPrint(("[EXT2]: Good we got one\n"));
						}
						else
						{
							KdPrint(("[EXT2]: This is not a directory!!!\n"));
							Status = STATUS_OBJECT_NAME_NOT_FOUND;
						}
					}
					if(Options & FILE_NON_DIRECTORY_FILE) 
					{
						KdPrint(("[EXT2]: We must open a file\n"));
						if(S_ISREG(INode->INode->i_mode))
						{
							KdPrint(("[EXT2]: Good we got one\n"));
						}
						else
						{
							KdPrint(("[EXT2]: This is not a file!!!?\n"));
							Status = STATUS_OBJECT_NAME_NOT_FOUND;
						}
					}
					

					XtUnLockINode(vcb, &INode);

					// only keep going if it is the correct type of object
					if(Status == STATUS_SUCCESS)
					{

						root_fcb = XtFSFindFCBbyINode(vcb, NextINode);

						if(root_fcb == 0)
						{
							// we need to allocate one
							root_fcb = XtFSAllocateFCB(vcb, NextINode);
						}

						root_ccb = XtFSAllocateCCB(root_fcb);
						//root_ccb = XtMMAllocatePoolStruct(NonPagedPool, sizeof(XtFSCCB), CCB);
						//root_ccb->fcb = root_fcb;

						// set the file name
						root_ccb->FileNameLen = Length;
						root_ccb->FileName = ExAllocatePool(NonPagedPool, root_ccb->FileNameLen);

						for(i = 0; i < Length; i++)
						{
							root_ccb->FileName[i] = Target[i];
						}

						if(root_ccb->FileNameLen == 1 && root_ccb->FileName[0] == '.')
						{
							KdPrint(("[EXT2]: The file name is .  Perhaps this should be set to the parent segment..."));
						}
            
						PtrIoStackLocation->FileObject->FsContext = &(root_fcb->CommonFCBHeader);
						PtrIoStackLocation->FileObject->FsContext2 = root_ccb;
						PtrIoStackLocation->FileObject->PrivateCacheMap = NULL;
						PtrIoStackLocation->FileObject->SectionObjectPointer = &(root_fcb->SectionObjectPointers);
						//ExInitializeResourceLite(&root_ccb->Resource);

						Info = FILE_OPENED;
						Status = STATUS_SUCCESS;
						KdPrint((DRIVER_NAME ": Create succeded\n"));
					}
				}
			}

			if(Status != STATUS_SUCCESS)
			{
				Status = STATUS_OBJECT_NAME_NOT_FOUND;
				Info = FILE_DOES_NOT_EXIST;
			}

			// free Target
			ExFreePool(Target);

		}
	}


	Irp->IoStatus.Status = Status;
	Irp->IoStatus.Information = Info;
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}

