/*
	$Header: /home/itig/cvsroot/ext2ifs/src/dirctl.c,v 1.11 2000/12/28 10:05:18 jn Exp $
*/

#include "ext2fs.h"

NTSTATUS StringPrint(PCH Msg, PSTRING String)
{
	int i;

	KdPrint((Msg));
	if(String)
	{
		for(i=0; i<String->Length; i++)
		{
			KdPrint(("%c", String->Buffer[i]));
		}
		KdPrint(("\n"));
	}
	else
	{
		KdPrint(("NULL\n"));
	}
	return STATUS_SUCCESS;
}


NTSTATUS
XtFSFillInDirectoryInfo(PFILE_DIRECTORY_INFORMATION Buffer, 
					  PULONG BufferLen,
						PtrXtFSCCB ccb)
{
	NTSTATUS Status;

	Buffer->FileNameLength = *BufferLen - sizeof(FILE_DIRECTORY_INFORMATION);
	Status = XtDirGetEntry(ccb->DirEnumObject, Buffer->FileName, &(Buffer->FileNameLength), &(Buffer->FileIndex));
	if(Status == STATUS_SUCCESS)
	{
		Buffer->NextEntryOffset = 0;
		Buffer->CreationTime.QuadPart = 0;
		Buffer->LastAccessTime.QuadPart = 0;
		Buffer->LastWriteTime.QuadPart = 0;
		Buffer->ChangeTime.QuadPart = 0;
		Buffer->EndOfFile.QuadPart = 1024;
		Buffer->AllocationSize.QuadPart = 1024;
		Buffer->FileAttributes = FILE_ATTRIBUTE_NORMAL | FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_DIRECTORY;
	}
	return Status;
}

NTSTATUS
XtFSFillInFullDirectoryInfo(PFILE_FULL_DIR_INFORMATION Buffer, 
					  PULONG BufferLen,
						PtrXtFSCCB ccb)
{
	NTSTATUS Status;

	Buffer->FileNameLength = *BufferLen - sizeof(FILE_FULL_DIR_INFORMATION);
	Status = XtDirGetEntry(ccb->DirEnumObject, Buffer->FileName, &(Buffer->FileNameLength), &(Buffer->FileIndex));
	if(Status == STATUS_SUCCESS)
	{
		Buffer->NextEntryOffset = 0;
		Buffer->CreationTime.QuadPart = 0;
		Buffer->LastAccessTime.QuadPart = 0;
		Buffer->LastWriteTime.QuadPart = 0;
		Buffer->ChangeTime.QuadPart = 0;
		Buffer->EndOfFile.QuadPart = 1024;
		Buffer->AllocationSize.QuadPart = 1024;
		Buffer->FileAttributes = FILE_ATTRIBUTE_NORMAL | FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_DIRECTORY;
		Buffer->EaSize = 0;
	}
	return Status;
}

NTSTATUS
XtFSFillInBothDirectoryInfo(PFILE_BOTH_DIR_INFORMATION Buffer, 
					  PULONG BufferLen,
						PtrXtFSCCB ccb)
{
	NTSTATUS Status;
	unsigned int i;

	Buffer->FileNameLength = *BufferLen - sizeof(FILE_BOTH_DIR_INFORMATION);
	Status = XtDirGetEntry(ccb->DirEnumObject, Buffer->FileName, &(Buffer->FileNameLength), &(Buffer->FileIndex));
	if(Status == STATUS_SUCCESS)
	{
		PINODE_OBJECT INode;
		XtLockINode(ccb->fcb->vcb, Buffer->FileIndex, &INode);

		// dump the file name...
		/*KdPrint(("[EXT2]: name len = %d\n", Buffer->FileNameLength));
		for(i = 0; i < Buffer->FileNameLength; i++)
		{
			KdPrint(("[EXT2]: %d = %c\n", i, Buffer->FileName[i]));
		}*/

		Buffer->NextEntryOffset = 0;
		XtUnixTimeToNTTime(&INode->INode->i_ctime, &Buffer->CreationTime);
		XtUnixTimeToNTTime(&INode->INode->i_atime, &Buffer->LastAccessTime);
		XtUnixTimeToNTTime(&INode->INode->i_mtime, &Buffer->LastWriteTime);
		XtUnixTimeToNTTime(&INode->INode->i_mtime, &Buffer->ChangeTime);
		Buffer->EndOfFile.QuadPart      = INode->INode->i_size;
		Buffer->AllocationSize.QuadPart = INode->INode->i_blocks * ccb->fcb->vcb->BLBlockSize;
		Buffer->FileAttributes = FILE_ATTRIBUTE_NORMAL | FILE_ATTRIBUTE_READONLY;// | FILE_ATTRIBUTE_DIRECTORY;
		KdPrint(("[EXT2]: format = %08x\n", INode->INode->i_mode & S_IFMT));
		if(S_ISDIR(INode->INode->i_mode))
		{
			Buffer->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
		}
		Buffer->EaSize = 0;
		Buffer->ShortNameLength = 0;

		XtUnLockINode(ccb->fcb->vcb, &INode);

		*BufferLen -= sizeof(FILE_BOTH_DIR_INFORMATION) - sizeof(WCHAR) + (sizeof(WCHAR) * Buffer->FileNameLength);
	}
	return Status;
}


NTSTATUS
XtFSFillInDirNameInfo(PFILE_NAMES_INFORMATION Buffer, 
					  PULONG BufferLen,
						PtrXtFSCCB ccb)
{
	NTSTATUS Status;

	Buffer->FileNameLength = *BufferLen - sizeof(FILE_NAMES_INFORMATION);
	Status = XtDirGetEntry(ccb->DirEnumObject, Buffer->FileName, &(Buffer->FileNameLength), &(Buffer->FileIndex));
	if(Status == STATUS_SUCCESS)
	{
		Buffer->NextEntryOffset = 0;
		*BufferLen -= sizeof(FILE_NAMES_INFORMATION) - sizeof(WCHAR) + (sizeof(WCHAR) * Buffer->FileNameLength);
	}
	return Status;
}



VOID
DumpDirectoryINode(PINODE_OBJECT INode,
  		  PtrXtFSVCB Vcb
);

NTSTATUS
XtFSDirectoryControl (
    IN PFSD_IRP_CONTEXT IrpContext
	)
{
    IN PIRP             Irp = IrpContext->Irp;
    PIO_STACK_LOCATION  IoStack;
	//PEXTENDED_IO_STACK_LOCATION IoStack;

	PFILE_OBJECT		FileObject;
	PVOID				Buffer;
	ULONG				BufferLen;
	PtrXtFSFCB			fcb;
	PtrXtFSCCB			ccb;
	PSTRING				Mask;

	NTSTATUS			status;

	ASSERT(Irp);


	if(IrpContext->DeviceObject == Global->DeviceObject)
	{
		DbgPrint(("[EXT2]: DirControl attept on driver device object\n"));
		status = STATUS_INVALID_PARAMETER;
		Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
		Irp->IoStatus.Information = 0;
	}
	else
	{
		IoStack = IoGetCurrentIrpStackLocation(Irp);

		ASSERT(IoStack);

		FileObject = IoStack->FileObject;

		ASSERT(FileObject);

		// get the ccb
		ccb = (PtrXtFSCCB)FileObject->FsContext2;

		// if the ccb is null then this is the vcb
		if(ccb == NULL)
		{
			status = STATUS_INVALID_PARAMETER;
		}
/*		else if(IrpContext->IsSynchronous == FALSE)
		{
			// we need to post this request...
			XtFSQueueRequest(IrpContext);
			status = STATUS_PENDING;
		}*/
		else
		{
			ASSERT(ccb);
			VERIFY(ccb, CCB);

			fcb = ccb->fcb;
			ASSERT(fcb);
			VERIFY(fcb, FCB);


			if(IoStack->MinorFunction == IRP_MN_QUERY_DIRECTORY)
			{
				KdPrint(("[EXT2]: Do directory listing\n"));
				// get the ccb lock...
				ACQUIRE_EX(ccb->Resource);
				BufferLen = ((PEXTENDED_IO_STACK_LOCATION)IoStack)->Parameters.QueryDirectory.Length;
				// we should make sure this is a directory...

				// prepare for the directory for scanning
				if(ccb->DirEnumObject)
				{
					if(IoStack->Flags & SL_RESTART_SCAN)
					{
						// free any stored state
						XtDestroyDirObject(&(ccb->DirEnumObject));
						ccb->DirEnumObject = 0;
						ccb->file_count = 0;
					}
					else
					{
						//KdPrint(("[EXT2]: enum object exists...\n"));
					}
				} 

				if(ccb->DirEnumObject == 0)
				{
					if(((PEXTENDED_IO_STACK_LOCATION)IoStack)->Parameters.QueryDirectory.FileName)
					{
						// create based on the supplied mask
						Mask = ((PEXTENDED_IO_STACK_LOCATION)IoStack)->Parameters.QueryDirectory.FileName;
						KdPrint(("[EXT2]: Mask len = %d [%c]\n", Mask->Length, Mask->Buffer[0]));
						if(Mask->Length == 2 && Mask->Buffer[0] == '*')
						{
							XtCreateDirObject(fcb->vcb, fcb->INode, &(ccb->DirEnumObject), 0);
						}
						else
						{
							StringPrint("[EXT2]: Create enum object for files matching ", ((PEXTENDED_IO_STACK_LOCATION)IoStack)->Parameters.QueryDirectory.FileName);
							XtCreateDirObject(fcb->vcb, fcb->INode, &(ccb->DirEnumObject), ((PEXTENDED_IO_STACK_LOCATION)IoStack)->Parameters.QueryDirectory.FileName);
						}
					}
					else
					{
						KdPrint(("[EXT2]: Create enum object for all files\n"));
						XtCreateDirObject(fcb->vcb, fcb->INode, &(ccb->DirEnumObject), 0);
					}
				}

				//
				Buffer = XtFSGetUserBuffer(Irp);

				//KdPrint(("[EXT2]: Buffer %08X len %08X\n", Buffer, BufferLen));
				//KdPrint(("[EXT2]: Mdl %08X\n", Irp->MdlAddress));
				//KdPrint(("[EXT2]: User Buffer %08X\n", Irp->UserBuffer));


				switch(((PEXTENDED_IO_STACK_LOCATION)IoStack)->Parameters.QueryDirectory.FileInformationClass)
				{
				case FileDirectoryInformation:
					KdPrint(("[EXT2]: FILE_DIRECTORY_INFORMATION\n"));
					status = XtFSFillInDirectoryInfo(Buffer, &BufferLen, ccb);
					break;

				case FileFullDirectoryInformation:
					KdPrint(("[EXT2]: FILE_FULL_DIRECTORY_INFORMATION\n"));
					status = XtFSFillInFullDirectoryInfo(Buffer, &BufferLen, ccb);
					break;

				case FileBothDirectoryInformation:
					KdPrint(("[EXT2]: FILE_BOTH_DIRECTORY_INFORMATION\n"));
					status = XtFSFillInBothDirectoryInfo(Buffer, &BufferLen, ccb);
					break;

				case FileNamesInformation:
					KdPrint(("[EXT2]: FILE_NAMES_INFORMATION\n"));
					status = XtFSFillInDirNameInfo(Buffer, &BufferLen, ccb);
					break;

				default:
					KdPrint(("[EXT2]: Unknown FileInformationClass\n"));
					status = STATUS_NOT_IMPLEMENTED;
				}
				//KdPrint(("[EXT2]: Status is %08x\n", status));

				if(status == STATUS_SUCCESS)
				{
					ccb->file_count++;
					XtMakeBlockPersistant(fcb->vcb, &(ccb->DirEnumObject->Block));
				}
				else
				{
					status = STATUS_NO_MORE_FILES;

					if(ccb->DirEnumObject)
					{
						XtDestroyDirObject(&(ccb->DirEnumObject));
					}
				}

				if(status == STATUS_NO_MORE_FILES && ccb->file_count == 0)
				{
					status = STATUS_NO_SUCH_FILE;
				}
				//KdPrint(("[EXT2]: Status now is %08x\n", status));

				RELEASE(ccb->Resource);
			}
			else
			{
				status = STATUS_NOT_IMPLEMENTED;
			}
		}

		Irp->IoStatus.Status = status;
		Irp->IoStatus.Information = 0;
	}
	if(status != STATUS_PENDING)
	{
		XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);
	}

	return status;
}
