/*
	$Header: /home/itig/cvsroot/ext2ifs/src/ext2fs.c,v 1.11 2001/01/22 07:38:28 jn Exp $

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Second Extended File System for Windows NT
	Written by John Newbigin

	All Functions prefixed with Xt

    XtFS - exported to Driver
	XtSB - superblock
	XtBG - block group
	XtBL - block
	XtIN - inode
	XtMB - bitmap
*/
#include "ext2fs.h"

XtFSGlobals *Global = NULL;

//#pragma code_seg("INIT")


/*
#define IRP_MJ_CREATE                   0x00
#define IRP_MJ_CREATE_NAMED_PIPE        0x01
#define IRP_MJ_CLOSE                    0x02
#define IRP_MJ_READ                     0x03
#define IRP_MJ_WRITE                    0x04
#define IRP_MJ_QUERY_INFORMATION        0x05
#define IRP_MJ_SET_INFORMATION          0x06
#define IRP_MJ_QUERY_EA                 0x07
#define IRP_MJ_SET_EA                   0x08
#define IRP_MJ_FLUSH_BUFFERS            0x09
#define IRP_MJ_QUERY_VOLUME_INFORMATION 0x0a
#define IRP_MJ_SET_VOLUME_INFORMATION   0x0b
#define IRP_MJ_DIRECTORY_CONTROL        0x0c
#define IRP_MJ_FILE_SYSTEM_CONTROL      0x0d
#define IRP_MJ_DEVICE_CONTROL           0x0e
#define IRP_MJ_INTERNAL_DEVICE_CONTROL  0x0f
#define IRP_MJ_SHUTDOWN                 0x10
#define IRP_MJ_LOCK_CONTROL             0x11
#define IRP_MJ_CLEANUP                  0x12
#define IRP_MJ_CREATE_MAILSLOT          0x13
#define IRP_MJ_QUERY_SECURITY           0x14
#define IRP_MJ_SET_SECURITY             0x15
#define IRP_MJ_QUERY_POWER              0x16
#define IRP_MJ_SET_POWER                0x17
#define IRP_MJ_DEVICE_CHANGE            0x18
#define IRP_MJ_QUERY_QUOTA              0x19
#define IRP_MJ_SET_QUOTA                0x1a
#define IRP_MJ_PNP_POWER                0x1b
#define IRP_MJ_MAXIMUM_FUNCTION         0x1b
*/

/*************************************************************************
*
* Function: DriverEntry()
*
* Description:
*	This routine is the standard entry point for all kernel mode drivers.
*	The routine is invoked at IRQL PASSIVE_LEVEL in the context of a
*	system worker thread.
*	All FSD specific data structures etc. are initialized here.
*
* Expected Interrupt Level (for execution) :
*
*  IRQL_PASSIVE_LEVEL
*
* Return Value: STATUS_SUCCESS/Error (will cause driver to be unloaded).
*
*************************************************************************/

NTSTATUS
DriverEntry (
	IN PDRIVER_OBJECT	DriverObject,
	IN PUNICODE_STRING	RegistryPath
	)
{
	UNICODE_STRING	DeviceName;
	NTSTATUS		status;
	PDEVICE_OBJECT	DeviceObject;

	/*HalDisplayString("\n");
	HalDisplayString("     o~o   Linux EXT2 Filesystem driver for Windows NT\n");
	HalDisplayString("     /V\\   Written by John Newbigin jn@it.swin.edu.au\n");
	HalDisplayString("    // \\\\  \n");
	HalDisplayString("   /| L |\\ http://uranus.it.swin.edu.au/~jn/linux\n");
	HalDisplayString("    ^`~'^  \n");
	HalDisplayString("\n");*/
	KdPrint(("[EXT2]: DriverEntry\n"));


	// set up dispatch functions
	DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_CREATE]              = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]               = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_READ]                = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_WRITE]               = XtFSBuildRequest;

    DriverObject->MajorFunction[IRP_MJ_CREATE_NAMED_PIPE]        = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION]        = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION]          = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_QUERY_EA]                 = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SET_EA]                   = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS]            = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_QUERY_VOLUME_INFORMATION] = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SET_VOLUME_INFORMATION]   = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL]        = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]           = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL]  = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SHUTDOWN]                 = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_LOCK_CONTROL]             = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP]                  = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_CREATE_MAILSLOT]          = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_QUERY_SECURITY]           = XtFSBuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SET_SECURITY]             = XtFSBuildRequest;



	DriverObject->DriverUnload = XtFSDriverUnload;

	// create a device
	RtlInitUnicodeString(&DeviceName, DEVICE_NAME);

	status = IoCreateDevice(
		DriverObject,
		sizeof(XtFSGlobals),
		&DeviceName,
		FILE_DEVICE_DISK_FILE_SYSTEM,
		0,
		FALSE,
		&DeviceObject
		);

	if (NT_SUCCESS(status))
	{
		IoRegisterFileSystem(DeviceObject);

		Global = (PtrXtFSGlobals)DeviceObject->DeviceExtension;

		Global->Identifier.Type = FGD;
		Global->Identifier.Size = sizeof(XtFSGlobals);

	    ExInitializeResourceLite(&Global->Resource);
	    InitializeListHead(&Global->NextVCB);

		Global->DriverObject = DriverObject;
		Global->DeviceObject = DeviceObject;
	}
	else
	{
		KdPrint(("[EXT2]: FileSystemControl: IoCreateDevice failed with error 0x%x\n", status));
	}

	//MmPageEntireDriver(DriverEntry);

    //
    //  Initialize the cache manager callback routines
    //

    Global->CacheManagerNoOpCallbacks.AcquireForLazyWrite  = &XtFSNoOpAcquire;
    Global->CacheManagerNoOpCallbacks.ReleaseFromLazyWrite = &XtFSNoOpRelease;
    Global->CacheManagerNoOpCallbacks.AcquireForReadAhead  = &XtFSNoOpAcquire;
    Global->CacheManagerNoOpCallbacks.ReleaseFromReadAhead = &XtFSNoOpRelease;


	//return status;
	return STATUS_SUCCESS;
}

//#pragma code_seg("PAGE")

VOID
XtFSDriverUnload (
	IN PDRIVER_OBJECT DriverObject
	)
{
	KdPrint(("[EXT2]: DriverUnload\n"));
}


NTSTATUS
XtFSBuildRequest (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    )
{
    BOOLEAN             IsTopLevelIrp = FALSE;
    PFSD_IRP_CONTEXT    IrpContext = NULL;
    NTSTATUS            Status = STATUS_UNSUCCESSFUL;

    try
    {
        try
        {
            FsRtlEnterFileSystem();

            if (!IoGetTopLevelIrp())
            {
                IsTopLevelIrp = TRUE;
                IoSetTopLevelIrp(Irp);
            }

            IrpContext = XtFSAllocateIrpContext(DeviceObject, Irp);

            if (!IrpContext)
            {
                Status = STATUS_INSUFFICIENT_RESOURCES;
                Irp->IoStatus.Status = Status;
                DbgIoCompleteRequest(Irp, IO_NO_INCREMENT);
            }
            else
            {
				//KdPrint(("[EXT2]: XtFSBuildRequest 0x%02x\n", IrpContext->MajorFunction));
                Status = XtFSDispatchRequest(IrpContext);
            }
        }
        except (XtFSExceptionFilter(IrpContext, GetExceptionInformation()))
        {
            Status = XtFSExceptionHandler(IrpContext, Irp);
        }
    }
    finally
    {
        if (IsTopLevelIrp)
        {
            IoSetTopLevelIrp(NULL);
        }

        FsRtlExitFileSystem();
    }
    
    return Status;
}

NTSTATUS
XtFSQueueRequest (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
	KdPrint(("[EXT2]: >> XtFSQueueRequest\n"));

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    // IsSynchronous means we can block (so we don't requeue it)
    IrpContext->IsSynchronous = TRUE;
	IrpContext->FreeContext = FALSE;
	IrpContext->Completed = TRUE;// This is to make it look like we have been completed
	IrpContext->Pending = TRUE;

    IoMarkIrpPending(IrpContext->Irp);

    ExInitializeWorkItem(
        &IrpContext->WorkQueueItem,
        XtFSDeQueueRequest,
        IrpContext
        );

    ExQueueWorkItem(&IrpContext->WorkQueueItem, CriticalWorkQueue);

    return STATUS_PENDING;
}

VOID
XtFSDeQueueRequest (
    IN PVOID Context
    )
{
    PFSD_IRP_CONTEXT IrpContext;

	KdPrint(("[EXT2]: << XtFSDeQueueRequest\n"));

    IrpContext = (PFSD_IRP_CONTEXT) Context;

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

	IrpContext->FreeContext = TRUE;
	IrpContext->Completed = FALSE;
    try
    {
        try
        {
            FsRtlEnterFileSystem();

            if (!IrpContext->IsTopLevel)
            {
                IoSetTopLevelIrp((PIRP) FSRTL_FSP_TOP_LEVEL_IRP);
            }

            XtFSDispatchRequest(IrpContext);

			if(IrpContext->FreeContext)
			{
				XtFSFreeIrpContext(IrpContext);
				KdPrint(("[EXT2]: XtFSDeQueueRequest2\n"));
			}
        }
        except (XtFSExceptionFilter(IrpContext, GetExceptionInformation()))
        {
            XtFSExceptionHandler(IrpContext, IrpContext->Irp);
        }
    }
    finally
    {
        IoSetTopLevelIrp(NULL);

        FsRtlExitFileSystem();
    }
	KdPrint(("[EXT2]: XtFSDeQueueRequest3\n"));
}


/*
	These functions are required to complete the IRP or ensure
	that the IRP is completed by another lower level component

	They must also free the IRP context when they are done with it.
*/
NTSTATUS
XtFSDispatchRequest (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
	NTSTATUS Status;
	KIRQL Irql;

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

	// check the IRQL
	Irql = KeGetCurrentIrql();
    //KdPrint((DRIVER_NAME ": XtFSDispatchRequest MJ = 0x%02x  IRQL = %d\n", IrpContext->MajorFunction, Irql));

    switch (IrpContext->MajorFunction)
    {
    case IRP_MJ_CREATE:
		Status = XtFSCreate(IrpContext);
		break;

    case IRP_MJ_CLOSE:
        Status = XtFSClose(IrpContext);
		break;

    case IRP_MJ_READ:
        Status = XtFSRead(IrpContext);
		break;

    case IRP_MJ_QUERY_INFORMATION:
        Status = XtFSQueryInformation(IrpContext);
		break;

/*    case IRP_MJ_SET_INFORMATION:
        return FsdSetInformation(IrpContext);

    //case IRP_MJ_QUERY_EA:

    case IRP_MJ_FLUSH_BUFFERS:
        return FsdFlushBuffers(IrpContext);
*/
    case IRP_MJ_QUERY_VOLUME_INFORMATION:
        Status = XtFSQueryVolumeInformation(IrpContext);
		break;

    case IRP_MJ_DIRECTORY_CONTROL:
        Status = XtFSDirectoryControl(IrpContext);
		break;

    case IRP_MJ_FILE_SYSTEM_CONTROL:
        Status = XtFSFileSystemControl(IrpContext);
		break;

    case IRP_MJ_DEVICE_CONTROL:
        Status = XtDeviceControl(IrpContext);
		break;

/*    case IRP_MJ_SHUTDOWN:
        return FsdShutdown(IrpContext);

    //case IRP_MJ_LOCK_CONTROL:
*/
    case IRP_MJ_CLEANUP:
        Status = XtFSCleanup(IrpContext);
		break;

    //case IRP_MJ_QUERY_SECURITY:
    //case IRP_MJ_QUERY_QUOTA:

#ifndef FSD_RO

    //case IRP_MJ_WRITE:
    //case IRP_MJ_SET_EA:
    //case IRP_MJ_SET_VOLUME_INFORMATION:
    //case IRP_MJ_SET_SECURITY:
    //case IRP_MJ_SET_QUOTA:

#endif // FSD_RO

    default:
        KdPrint((DRIVER_NAME ": XtFSDispatchRequest: Unknown MajorFunction 0x%x\n",
            IrpContext->MajorFunction
            ));

        IrpContext->Irp->IoStatus.Status = STATUS_NOT_SUPPORTED;//STATUS_DRIVER_INTERNAL_ERROR;
        XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

        Status = STATUS_NOT_SUPPORTED;//STATUS_DRIVER_INTERNAL_ERROR;
    }
	if(IrpContext->Completed == FALSE)
	{
        KdPrint((DRIVER_NAME ": The IRP does not seem to be completed!\n"));
	}
	if((IrpContext->Pending == FALSE) && (IrpContext->FreeContext == TRUE))
	{
        XtFSFreeIrpContext(IrpContext);
	}
	return Status;
}

NTSTATUS
XtFSExceptionFilter (
    IN PFSD_IRP_CONTEXT     IrpContext,
    IN PEXCEPTION_POINTERS  ExceptionPointers
    )
{
    NTSTATUS ExceptionCode;
    NTSTATUS Status;

    //
    // Only use a valid IrpContext
    //
    if (IrpContext)
    {
        if ((IrpContext->Identifier.Type != ICX) ||
            (IrpContext->Identifier.Size != sizeof(FSD_IRP_CONTEXT)))
        {
            IrpContext = NULL;
        }
    }

    ExceptionCode = ExceptionPointers->ExceptionRecord->ExceptionCode;

    if ((ExceptionCode == STATUS_IN_PAGE_ERROR) &&
        (ExceptionPointers->ExceptionRecord->NumberParameters >= 3))
    {
        ExceptionCode =
            ExceptionPointers->ExceptionRecord->ExceptionInformation[2];
    }

    //
    // If the exception is expected execute our handler
    //
    if (FsRtlIsNtstatusExpected(ExceptionCode))
    {
        KdPrint((DRIVER_NAME ": XtFSExceptionFilter: Catching exception 0x%x\n",
            ExceptionCode));

        Status = EXCEPTION_EXECUTE_HANDLER;

        if (IrpContext)
        {
            IrpContext->ExceptionInProgress = TRUE;
            IrpContext->ExceptionCode = ExceptionCode;
        }
    }
    //
    // else continue search for an higher level exception handler
    //
    else
    {
        KdPrint((DRIVER_NAME ": XtFSExceptionFilter: Passing on exception 0x%x\n",
            ExceptionCode));

        Status = EXCEPTION_EXECUTE_HANDLER;//EXCEPTION_CONTINUE_SEARCH; // for testing
        KdPrint((DRIVER_NAME ": -- Letting exception go --\n"));

        if (IrpContext)
        {
            XtFSFreeIrpContext(IrpContext);
        }
    }

    return Status;
}

NTSTATUS
XtFSExceptionHandler (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PIRP             Irp
    )
{
    NTSTATUS Status;

    //
    // Complete the IRP with the exception code as status
    //

    if (IrpContext)
    {
        Status = IrpContext->ExceptionCode;
        //XtFSFreeIrpContext(IrpContext);
    }
    else
    {
        Status = STATUS_INSUFFICIENT_RESOURCES;
    }

    Irp->IoStatus.Status = Status;
    XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

    return Status;
}

/*VOID
FsdCharToWchar (
    IN OUT PWCHAR   Destination,
    IN PCHAR        Source,
    IN ULONG        Length
    )
{
    ULONG Index;

    ASSERT(Destination != NULL);
    ASSERT(Source != NULL);

    for (Index = 0; Index < Length; Index++)
    {
        Destination[Index] = (WCHAR) Source[Index];
    }
}*/


/*NTSTATUS
FsdWcharToChar (
    IN OUT PCHAR    Destination,
    IN PWCHAR       Source,
    IN ULONG        Length
    )
{
    ULONG       Index;
    NTSTATUS    Status;

    ASSERT(Destination != NULL);
    ASSERT(Source != NULL);

    for (Index = 0, Status = STATUS_SUCCESS; Index < Length; Index++)
    {
        Destination[Index] = (CHAR) Source[Index];

        //
        // Check that the wide character fits in a normal character
        // but continue with the conversion anyway
        //
        if ( ((WCHAR) Destination[Index]) != Source[Index])
        {
            KdPrint((DRIVER_NAME ": FsdWcharToChar: Invalid character: %C\n",
                Source[Index]));

            Status = STATUS_OBJECT_NAME_INVALID;
        }
    }

    return Status;
}*/


NTSTATUS
XtFSLockUserBuffer (
    IN PIRP             Irp,
    IN ULONG            Length,
    IN LOCK_OPERATION   Operation
    )
{
    NTSTATUS Status;

    ASSERT(Irp != NULL);

    if (Irp->MdlAddress != NULL)
    {
        return STATUS_SUCCESS;
    }

    IoAllocateMdl(Irp->UserBuffer, Length, FALSE, FALSE, Irp);

    if (Irp->MdlAddress == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    __try
    {
        MmProbeAndLockPages(Irp->MdlAddress, Irp->RequestorMode, Operation);

        Status = STATUS_SUCCESS;
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        KdPrint((DRIVER_NAME ": XtFSLockUserBuffer: Invalid User Buffer\n"));

        IoFreeMdl(Irp->MdlAddress);

        Irp->MdlAddress = NULL;

        Status = STATUS_INVALID_USER_BUFFER;
    }

    return Status;
}

PVOID
XtFSGetUserBuffer (
    IN PIRP Irp
    )
{
    ASSERT(Irp != NULL);

    if (Irp->MdlAddress != NULL)
    {
        return MmGetSystemAddressForMdl(Irp->MdlAddress);
    }
    else
    {
        return Irp->UserBuffer;
    }
}


PFSD_IRP_CONTEXT
XtFSAllocateIrpContext (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    )
{
    PIO_STACK_LOCATION  IoStackLocation;
    PFSD_IRP_CONTEXT    IrpContext;

    ASSERT(DeviceObject != NULL);
    ASSERT(Irp != NULL);

    IoStackLocation = IoGetCurrentIrpStackLocation(Irp);

    //
    // Allocate an IrpContext
    //

    IrpContext = XtMMAllocatePool(NonPagedPool, sizeof(FSD_IRP_CONTEXT));

    if (!IrpContext)
    {
        return NULL;
    }

    //
    // Initialize the IrpContext with information about the request
    //

    IrpContext->Identifier.Type = ICX;
    IrpContext->Identifier.Size = sizeof(FSD_IRP_CONTEXT);

    IrpContext->Irp = Irp;

    IrpContext->MajorFunction = IoStackLocation->MajorFunction;
    IrpContext->MinorFunction = IoStackLocation->MinorFunction;

    IrpContext->DeviceObject = DeviceObject;

    IrpContext->FileObject = IoStackLocation->FileObject;

    if (IrpContext->MajorFunction == IRP_MJ_FILE_SYSTEM_CONTROL ||
        IrpContext->MajorFunction == IRP_MJ_DEVICE_CONTROL ||
        IrpContext->MajorFunction == IRP_MJ_SHUTDOWN)
    {
        IrpContext->IsSynchronous = TRUE;
    }
    else if (IrpContext->MajorFunction == IRP_MJ_CLEANUP)
    {
        IrpContext->IsSynchronous = FALSE;
    }
    else
    {
        IrpContext->IsSynchronous = IoIsOperationSynchronous(Irp);
    }

    IrpContext->IsTopLevel = (IoGetTopLevelIrp() == Irp);

    IrpContext->ExceptionInProgress = FALSE;

	IrpContext->Completed = FALSE;
	IrpContext->FreeContext = TRUE;
	IrpContext->Pending = FALSE;

    return IrpContext;
}

VOID
XtFSFreeIrpContext (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    ASSERT(IrpContext != NULL);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    //KdPrint((DRIVER_NAME ": Freeing IrpContext\n"));

    XtMMFreePool(IrpContext, sizeof(FSD_IRP_CONTEXT));
}


VOID 
XtFSCompleteRequest(
		IN PFSD_IRP_CONTEXT  IrpContext,
		IN CCHAR  PriorityBoost
		)
{
	if(IrpContext->Completed)
	{
        KdPrint((DRIVER_NAME ": IRP has already been completed\n"));
	}
	else
	{
		IrpContext->Completed = TRUE;
		IoCompleteRequest(IrpContext->Irp, PriorityBoost);
	}
}
