/*
	$Header: /home/itig/cvsroot/ext2ifs/src/fileinfo.c,v 1.4 2000/11/07 11:46:47 jn Exp $
*/

#include "ext2fs.h"

NTSTATUS
XtFSFillInBasicInfo(PtrXtFSCCB ccb,
					PFILE_BASIC_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_BASIC_INFORMATION) );

	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_ctime, &Buffer->CreationTime);
	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_atime, &Buffer->LastAccessTime);
	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_mtime, &Buffer->LastWriteTime);
	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_mtime, &Buffer->ChangeTime);

	Buffer->FileAttributes = FILE_ATTRIBUTE_NORMAL;// | FILE_ATTRIBUTE_DIRECTORY;
	if(S_ISDIR(ccb->fcb->INode->INode->i_mode))
	{
		Buffer->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
	}

	*BufferLen -= sizeof(FILE_BASIC_INFORMATION);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSFillInStandardInfo(PtrXtFSCCB ccb,
					PFILE_STANDARD_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_STANDARD_INFORMATION) );

	Buffer->AllocationSize.QuadPart = ccb->fcb->INode->INode->i_size;
	Buffer->EndOfFile.QuadPart = ccb->fcb->INode->INode->i_size;
	Buffer->NumberOfLinks = ccb->fcb->INode->INode->i_links_count;
	Buffer->DeletePending = FALSE;

	if(S_ISDIR(ccb->fcb->INode->INode->i_mode))
	{
		Buffer->Directory = TRUE;
	}
	else
	{
		Buffer->Directory = FALSE;
	}

	*BufferLen -= sizeof(FILE_STANDARD_INFORMATION);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSFillInNetowrkInfo(PtrXtFSCCB ccb,
					PFILE_NETWORK_OPEN_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_NETWORK_OPEN_INFORMATION) );

	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_ctime, &Buffer->CreationTime);
	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_atime, &Buffer->LastAccessTime);
	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_mtime, &Buffer->LastWriteTime);
	XtUnixTimeToNTTime(&ccb->fcb->INode->INode->i_mtime, &Buffer->ChangeTime);

	Buffer->AllocationSize.QuadPart = ccb->fcb->INode->INode->i_size;
	Buffer->EndOfFile.QuadPart = ccb->fcb->INode->INode->i_size;

	Buffer->FileAttributes = FILE_ATTRIBUTE_NORMAL;// | FILE_ATTRIBUTE_DIRECTORY;
	if(S_ISDIR(ccb->fcb->INode->INode->i_mode))
	{
		Buffer->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
	}

	*BufferLen -= sizeof(FILE_NETWORK_OPEN_INFORMATION);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSFillInInternalInfo(PtrXtFSCCB ccb,
					PFILE_INTERNAL_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_INTERNAL_INFORMATION) );

	Buffer->IndexNumber.QuadPart = ccb->fcb->INode->INodeNo;

	*BufferLen -= sizeof(FILE_INTERNAL_INFORMATION);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSFillInExtendedInfo(PtrXtFSCCB ccb,
					PFILE_EA_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_EA_INFORMATION) );

	Buffer->EaSize = 0;

	*BufferLen -= sizeof(FILE_EA_INFORMATION);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSFillInNameInfo(PtrXtFSCCB ccb,
					PFILE_NAME_INFORMATION Buffer,
					PLONG BufferLen)
{
	ULONG i;
	NTSTATUS Status = STATUS_SUCCESS;

    RtlZeroMemory( Buffer, sizeof(FILE_NAME_INFORMATION) );

	if(ccb->FileName)
	{
		Buffer->FileNameLength = ccb->FileNameLen * sizeof(WCHAR);
		if((*BufferLen) >= (LONG)(sizeof(FILE_NAME_INFORMATION) - sizeof(WCHAR) + Buffer->FileNameLength))
		{
			for(i = 0; i < ccb->FileNameLen; i++)
			{
				Buffer->FileName[i] = (WCHAR)ccb->FileName[i];
			}
		}
		else
		{
			Status = STATUS_BUFFER_OVERFLOW;
		}
	}
	else
	{
		// assume this is the root...
		Buffer->FileNameLength = 2;
		Buffer->FileName[0] = '\\';
	}

	*BufferLen -= sizeof(FILE_NAME_INFORMATION) - sizeof(WCHAR) + Buffer->FileNameLength;

	return Status;
}

/*NTSTATUS
XtFSFillInCompressionInfo(PtrXtFSCCB ccb,
					PFILE_COMPRESSION_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_COMPRESSION_INFORMATION) );

	Buffer->CompressedFileSize = ccb->fcb->INode->INode->i_size;

	*BufferLen -= sizeof(FILE_COMPRESSION_INFORMATION);

	return STATUS_SUCCESS;
}*/

NTSTATUS
XtFSFillInPositionInfo(PtrXtFSCCB ccb,
					PFILE_POSITION_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_POSITION_INFORMATION) );

	Buffer->CurrentByteOffset.QuadPart = 0;

	*BufferLen -= sizeof(FILE_POSITION_INFORMATION);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSFillInAllInfo(PtrXtFSCCB ccb,
					PFILE_ALL_INFORMATION Buffer,
					PLONG BufferLen)
{
    RtlZeroMemory( Buffer, sizeof(FILE_ALL_INFORMATION) );

	XtFSFillInBasicInfo(ccb, &(Buffer->BasicInformation), BufferLen);
	XtFSFillInStandardInfo(ccb, &(Buffer->StandardInformation), BufferLen);
	XtFSFillInInternalInfo(ccb, &(Buffer->InternalInformation), BufferLen);
	XtFSFillInExtendedInfo(ccb, &(Buffer->EaInformation), BufferLen);
	//XtFSFillInAccessInfo(ccb, &(Buffer->AccessInformation), BufferLen);
	XtFSFillInPositionInfo(ccb, &(Buffer->PositionInformation), BufferLen);
	//XtFSFillInModeInfo(ccb, &(Buffer->ModeInformation), BufferLen);
	//XtFSFillInAlignmentInfo(ccb, &(Buffer->AlignmentInformation), BufferLen);
	XtFSFillInNameInfo(ccb, &(Buffer->NameInformation), BufferLen);

	return STATUS_SUCCESS;
}

NTSTATUS
XtFSQueryInformation (
    IN PFSD_IRP_CONTEXT IrpContext
	)
{
    IN PIRP             Irp = IrpContext->Irp;
    PIO_STACK_LOCATION  IoStack;
	PFILE_OBJECT		FileObject;
	PVOID				Buffer;
	ULONG				BufferLen;
	PtrXtFSFCB			fcb;
	PtrXtFSCCB			ccb;

	NTSTATUS			status;

	ASSERT(Irp);

	if(IrpContext->DeviceObject == Global->DeviceObject)
	{
		KdPrint(("[EXT2]: FileInfo attept on driver device object\n"));

		status = STATUS_INVALID_PARAMETER;
	}
	else
	{
     
		IoStack = IoGetCurrentIrpStackLocation(Irp);

		ASSERT(IoStack);

		FileObject = IoStack->FileObject;

		ASSERT(FileObject);

		// get the ccb
		ccb = (PtrXtFSCCB)FileObject->FsContext2;

		// if the ccb is null then this is the device object...
		if(ccb == NULL)
		{
			status = STATUS_INVALID_PARAMETER;
		}
		else if(IrpContext->IsSynchronous == FALSE)
		{
			KdPrint(("[EXT2]: FileInfo is not synchronous so posting...\n"));
			// we need to post this request...
			XtFSQueueRequest(IrpContext);
			status = STATUS_PENDING;
		}
		else
		{

			ASSERT(ccb);
			VERIFY(ccb, CCB);

			fcb = ccb->fcb;
			ASSERT(fcb);
			VERIFY(fcb, FCB);

			Buffer = Irp->AssociatedIrp.SystemBuffer;
			BufferLen = IoStack->Parameters.QueryFile.Length;

//			KdPrint(("[EXT2]: Buffer %08X len %08X\n", Buffer, BufferLen));
			KdPrint(("[EXT2]: Do file info on inode %d\n", fcb->inode_number));


			switch(IoStack->Parameters.QueryFile.FileInformationClass)
			{
			case FileBasicInformation:
				KdPrint(("[EXT2]: FILE_BASIC_INFORMATION\n"));
				status = XtFSFillInBasicInfo(ccb, Buffer, &BufferLen);
				break;

			case FileStandardInformation:
				KdPrint(("[EXT2]: FILE_STANDARD_INFORMATION\n"));
				status = XtFSFillInStandardInfo(ccb, Buffer, &BufferLen);
				break;

			case FileNetworkOpenInformation:
				KdPrint(("[EXT2]: FILE_NETWORK_OPEN_INFORMATION\n"));
				status = XtFSFillInNetowrkInfo(ccb, Buffer, &BufferLen);
				break;

			case FileInternalInformation:
				KdPrint(("[EXT2]: FILE_INTERNAL_INFORMATION\n"));
				status = XtFSFillInInternalInfo(ccb, Buffer, &BufferLen);
				break;

			case FileEaInformation:
				KdPrint(("[EXT2]: FILE_EA_INFORMATION\n"));
				status = XtFSFillInExtendedInfo(ccb, Buffer, &BufferLen);
				break;

			case FileNameInformation:
				KdPrint(("[EXT2]: FILE_NAME_INFORMATION\n"));
				status = XtFSFillInNameInfo(ccb, Buffer, &BufferLen);
				break;

			case FileCompressionInformation:
				KdPrint(("[EXT2]: FILE_COMPRESSION_INFORMATION\n"));
				status = STATUS_NOT_IMPLEMENTED;
				break;

			case FilePositionInformation:
				KdPrint(("[EXT2]: FILE_POSITION_INFORMATION\n"));
				status = XtFSFillInPositionInfo(ccb, Buffer, &BufferLen);
				break;

			case FileAllInformation:
				KdPrint(("[EXT2]: FILE_ALL_INFORMATION\n"));
				status = XtFSFillInAllInfo(ccb, Buffer, &BufferLen);
				break;

			case FileStreamInformation:
				KdPrint(("[EXT2]: FILE_STREAM_INFORMATION\n"));
				status = STATUS_NOT_IMPLEMENTED;
				break;

			default:
				KdPrint(("[EXT2]: Unknown FileInformationClass\n"));
				status = STATUS_NOT_IMPLEMENTED;
			}
		}
	}

	Irp->IoStatus.Status = status;
	Irp->IoStatus.Information = IoStack->Parameters.QueryFile.Length - BufferLen;  // actual used ammount of buffer
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return status;
}
