/*
	$Header: /home/itig/cvsroot/ext2ifs/src/read.c,v 1.13 2001/07/14 11:54:10 jn Exp $
*/

#include "ext2fs.h"
NTSTATUS
FsdReadBlockDevice (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PLARGE_INTEGER   Offset,
    IN ULONG            Length,
    IN OUT PVOID        Buffer
    );



NTSTATUS
XtFSRead (
    IN PFSD_IRP_CONTEXT IrpContext
	)
{
    IN PIRP             Irp = IrpContext->Irp;
	NTSTATUS			Status = STATUS_NOT_IMPLEMENTED;
    PIO_STACK_LOCATION  IoStack;
	PFILE_OBJECT		FileObject;
	PtrXtFSVCB			Vcb;
	PtrXtFSCCB			Ccb;

    ULONG               Length;
    LARGE_INTEGER       ByteOffset;
    PUCHAR              Buffer;

	ASSERT(Irp);
        
    IoStack = IoGetCurrentIrpStackLocation(Irp);

	ASSERT(IoStack);

	DbgPrint("[EXT2]: Read minor function = %d\n", IoStack->MinorFunction);

	FileObject = IoStack->FileObject;

	ASSERT(FileObject);
	ASSERT(Global);

	// see if it is the volume it's self that we have to read...
	if(IrpContext->DeviceObject == Global->DeviceObject)
	{
		DbgPrint(("[EXT2]: Read attept on driver device object\n"));
		Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
		Irp->IoStatus.Information = 0;
		XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);
		Status = STATUS_INVALID_PARAMETER;
	}
	else if(FileObject->FsContext2 == NULL)
	{
		//DbgPrint("[EXT2]: Read of device?\n");
		if(FileObject->FsContext == NULL)
		{
			Vcb = (PtrXtFSVCB)IrpContext->DeviceObject->DeviceExtension;
			ASSERT(Vcb);
			VERIFY(Vcb, VCB);

			Length = IoStack->Parameters.Read.Length;
			ByteOffset = IoStack->Parameters.Read.ByteOffset;
			Buffer = XtFSGetUserBuffer(Irp);

			KdPrint((DRIVER_NAME ": Reading block device directly\n"));
			// create a new read request...?
			if(Buffer == NULL)
			{
				Status = STATUS_INVALID_USER_BUFFER;
			}
			else
			{
				Status = XtBLReadBlockDevice(Vcb->TargetDeviceObject, 
									&ByteOffset,
									Length, 
									Buffer);
			}

			Irp->IoStatus.Status = Status;//STATUS_INVALID_PARAMETER;
			Irp->IoStatus.Information = 0;
			if(NT_SUCCESS(Status))
			{
				Irp->IoStatus.Information = Length;
			}
			XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);
			//Status = STATUS_INVALID_PARAMETER;
		}
		else
		{
			ASSERT(FileObject->FsContext);
			// at the moment it must be???
			Vcb = CONTAINING_RECORD(FileObject->FsContext, XtFSVCB, CommonFCBHeader);
			ASSERT(Vcb);
			VERIFY(Vcb, VCB);

			Length = IoStack->Parameters.Read.Length;
			ByteOffset = IoStack->Parameters.Read.ByteOffset;

			//DbgPrint("[EXT2]: Getting buffer\n");
			Buffer = XtFSGetUserBuffer(Irp);

			if (Buffer == NULL)
			{
				KdPrint((DRIVER_NAME ": FsdRead: Invalid User Buffer\n"));
				Status = STATUS_INVALID_USER_BUFFER;
				Irp->IoStatus.Status = STATUS_INVALID_USER_BUFFER;
				Irp->IoStatus.Information = 0;
				XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);
			}
			else
			{
				//KdPrint((DRIVER_NAME": Irp->CurrentLocation = %d, StackSize = %d\n", Irp->CurrentLocation, IrpContext->DeviceObject->StackSize));
				//KdPrint((DRIVER_NAME ": Ready to read\n"));
				Status = XtBLReadBlockDeviceSync(
					IrpContext,
					Vcb,
					&ByteOffset,
					Length,
					Buffer
					);
				//Status = FsdReadBlockDevice(Vcb->TargetDeviceObject, &ByteOffset, Length, Buffer);
			}
			//DbgPrint("[EXT2]: Read\n");
		}
	}
	else
	{
		// read a file

		DbgPrint(("[EXT2]: Starting file read\n"));

		ASSERT(FileObject->FsContext);

		Ccb = FileObject->FsContext2;
		ASSERT(Ccb);
		VERIFY(Ccb, CCB);

		Vcb = Ccb->fcb->vcb;
		ASSERT(Vcb);
		VERIFY(Vcb, VCB);

		Length = IoStack->Parameters.Read.Length;
		ByteOffset = IoStack->Parameters.Read.ByteOffset;

		Buffer = XtFSGetUserBuffer(Irp);

		if (Buffer == NULL)
		{
			KdPrint((DRIVER_NAME ": FsdRead: Invalid User Buffer for file read\n"));
			Status = STATUS_INVALID_USER_BUFFER;
			Irp->IoStatus.Status = STATUS_INVALID_USER_BUFFER;
			Irp->IoStatus.Information = 0;
			XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);
		}
		else
		{
			Status = XtBLReadFile(
				IrpContext,
				Vcb,
				Ccb,
				&ByteOffset,
				Length,
				Buffer
				);

		}

	}

	return Status;
}
