/*
	$Header: /home/itig/cvsroot/ext2ifs/src/volinfo.c,v 1.6 2000/10/26 11:33:50 jn Exp $
*/

#include "ext2fs.h"

NTSTATUS
XtQueryFsSizeInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_SIZE_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    Buffer->TotalAllocationUnits.QuadPart = Vcb->SuperBlock->s_blocks_count;
    Buffer->AvailableAllocationUnits.QuadPart = Vcb->SuperBlock->s_free_blocks_count;
    Buffer->SectorsPerAllocationUnit = Vcb->BLBlockSize / Vcb->DiskGeometry.BytesPerSector;
    Buffer->BytesPerSector = Vcb->DiskGeometry.BytesPerSector;

    *Length -= sizeof( FILE_FS_SIZE_INFORMATION );

    return STATUS_SUCCESS;
}

NTSTATUS
XtQueryFsVolumeInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_VOLUME_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    NTSTATUS Status = STATUS_SUCCESS;

	Buffer->VolumeCreationTime.QuadPart = 0;
	Buffer->VolumeSerialNumber = 0;
	Buffer->SupportsObjects = FALSE;

	Buffer->VolumeLabelLength = 0;

    *Length -= sizeof( PFILE_FS_VOLUME_INFORMATION );

	return Status;
}

NTSTATUS
XtQueryFsDeviceInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_DEVICE_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    Buffer->Characteristics = Vcb->TargetDeviceObject->Characteristics;
    Buffer->DeviceType = FILE_DEVICE_DISK;

    *Length -= sizeof( FILE_FS_DEVICE_INFORMATION );

    return STATUS_SUCCESS;
}


NTSTATUS
XtQueryFsAttributeInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_ATTRIBUTE_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    NTSTATUS Status = STATUS_SUCCESS;
	WCHAR *DriverName = L"EXT2"; // len = 8
    ULONG BytesToCopy;

	Buffer->FileSystemAttributes = FILE_CASE_SENSITIVE_SEARCH | FILE_CASE_PRESERVED_NAMES;
	Buffer->MaximumComponentNameLength = EXT2_NAME_LEN;

	*Length = sizeof(FILE_FS_ATTRIBUTE_INFORMATION) - sizeof(WCHAR);

	if(*Length >= 8)
	{
		// Enough room for the name
		BytesToCopy = 8;
	}
	else
	{
		BytesToCopy = *Length;
		Status = STATUS_BUFFER_OVERFLOW;
	}

	Buffer->FileSystemNameLength = BytesToCopy;

	RtlCopyMemory(
			Buffer->FileSystemName,
			DriverName,
			BytesToCopy	);

	*Length -= BytesToCopy;

	return Status;
}

NTSTATUS
XtFSQueryVolumeInformation (
    IN PFSD_IRP_CONTEXT IrpContext
	)
{
    IN PIRP             Irp = IrpContext->Irp;
    PIO_STACK_LOCATION  IoStack;
	PVOID				Buffer;
	ULONG				BufferLen;
	PFCB				Fcb;
	PCCB				Ccb;

	NTSTATUS			Status = STATUS_INVALID_PARAMETER;


	ASSERT(Irp);
        
    IoStack = IoGetCurrentIrpStackLocation(Irp);

	ASSERT(IoStack);

	// make sure this is not the VCB???
	if(IoStack->FileObject->FsContext2 == NULL)
	{
		KdPrint(("[ETX2]: XtFSQueryVolumeInformationon on VCB\n"));
		Status = STATUS_INVALID_PARAMETER;
	}
	else
	{
		Ccb = (PtrXtFSCCB)IoStack->FileObject->FsContext2;
		Fcb = Ccb->fcb;

		Buffer = Irp->AssociatedIrp.SystemBuffer;
		BufferLen = IoStack->Parameters.QueryVolume.Length;

		RtlZeroMemory(Buffer, BufferLen);

		switch(IoStack->Parameters.QueryVolume.FsInformationClass)
		{
			case FileFsVolumeInformation:
				KdPrint(("[ETX2]: FileFsVolumeInformation\n"));
				Status = XtQueryFsSizeInfo( IrpContext, Fcb->vcb, Buffer, &BufferLen );
				break;

			case FileFsSizeInformation:
				KdPrint(("[ETX2]: FileFsSizeInformation\n"));
	            Status = XtQueryFsSizeInfo( IrpContext, Fcb->vcb, Buffer, &BufferLen );
				break;

			case FileFsDeviceInformation:
				KdPrint(("[ETX2]: FileFsDeviceInformation\n"));
	            Status = XtQueryFsDeviceInfo( IrpContext, Fcb->vcb, Buffer, &BufferLen );
				break;

			case FileFsAttributeInformation:
				KdPrint(("[ETX2]: FileFsAttributeInformation\n"));
	            Status = XtQueryFsAttributeInfo( IrpContext, Fcb->vcb, Buffer, &BufferLen );
				break;

			default:
				KdPrint((DRIVER_NAME ": FsdQueryVolumeInformation: "
					"Unknown FsInformationClass %d\n", IoStack->Parameters.QueryVolume.FsInformationClass));

				Status = STATUS_INVALID_PARAMETER;

		}
		Irp->IoStatus.Information = IoStack->Parameters.QueryVolume.Length - BufferLen;
	}


	Irp->IoStatus.Status = Status;
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}
